#pragma once

/// Indicates the major version of LibTorch.
#define TORCH_VERSION_MAJOR 1

/// Indicates the minor version of LibTorch.
#define TORCH_VERSION_MINOR 10

/// Indicates the patch version of LibTorch.
#define TORCH_VERSION_PATCH 2
